#
# Product Manager
#
# (C) Copyright 2000,2001 The Internet (Aust) Pty Ltd
# ACN: 082 081 472  ABN: 83 082 081 472
# All Rights Reserved
#
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
# OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
# OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
# SUCH DAMAGE.
#
# Author: Andrew Milton <akm@theinternet.com.au>

import os
import urllib
import time
import string
import re

from Globals import DTMLFile
from OFS.Folder import Folder
from OFS.Application import get_products, install_product, get_folder_permissions

from ProductPackage import ProductPackageMetadata, PackageListMetadata, metadataToProductPackage
from FormWizard.FormWizard import FormWizard
from BTrees.OOBTree import OOBTree
from ProductInstaller.ProductInstaller import ProductInstaller
from ZODB.ExportImport import TemporaryFile
from Products.ZSQLMethods.SQL import SQL
from Shared.DC.ZRDB import Results
from string import upper, lower
import Missing
mt=type(Missing.Value)

def manage_addProductManager(self, id, title='', connection='',
							 REQUEST={}):
	""" Add a Product Manager """

	if connection=='zodb':
		o = ProductManager(id, title)
	else:
		o = ProductManagerRDBMS(id, connection, title)

	self._setObject(id, o)

	if REQUEST:
		return self.manage_main(self, REQUEST)
	return ''

manage_addProductManagerForm=DTMLFile('dtml/manage_addProductManagerForm', globals())

class ProductManager(Folder):
	""" A Product Management Product """

	meta_type='Product Manager'

	manage_createProductWizard=FormWizard(('manage_createProductForm',
										   'manage_getProductArchitectureForm',
										   'manage_getProductDatabaseRequirements',
										   'manage_classifyProductForm',
										   'manage_addDownloadSitesForm',
										   'manage_getPackageDependencies',),
										  'manage_createProduct',
										  'dtml')

	manage_viewPackageList=DTMLFile('dtml/manage_viewPackageList', globals())
	viewPackage=DTMLFile('dtml/viewPackage', globals())
	viewCategory=DTMLFile('dtml/viewCategory', globals())
	manage_importPackageListForm=DTMLFile('dtml/manage_importPackageListForm', globals())

	manage_licensesForm=DTMLFile('dtml/manage_licensesForm', globals())
	manage_categoriesForm=DTMLFile('dtml/manage_categoriesForm', globals())
	manage_statusesForm=DTMLFile('dtml/manage_statusForm', globals())
	manage_databaseForm=DTMLFile('dtml/manage_databaseForm', globals())
	manage_operatingSystemsForm=DTMLFile('dtml/manage_operatingSystemsForm', globals())
	manage_CPUForm=DTMLFile('dtml/manage_CPUForm', globals())

	
	manage_exportPackageListForm=DTMLFile('dtml/manage_exportPackageListForm', globals())

	manage_updateForm=DTMLFile('dtml/manage_updateForm', globals())
	manage_editPackageForm=DTMLFile('dtml/manage_editPackageForm', globals())
	manage_editArchitectureForm=DTMLFile('dtml/manage_editArchitectureForm', globals())
	manage_editDatabaseRequirementsForm=DTMLFile('dtml/manage_editDatabaseRequirementsForm', globals())
	manage_editClassificationForm=DTMLFile('dtml/manage_editClassificationForm', globals())
	manage_editDownloadSitesForm=DTMLFile('dtml/manage_editDownloadSitesForm', globals())
	manage_editDependenciesForm=DTMLFile('dtml/manage_editDependenciesForm', globals())
	
	manage_options=(
		{'label':'Package List', 'action':'manage_viewPackageList'},
		{'label':'New Package', 'action':'manage_createProductWizard'},
		{'label':'Export PackageList', 'action':'manage_exportPackageListForm'},
		{'label':'Import Packagelist', 'action':'manage_importPackageListForm'},
		{'label':'Licenses', 'action':'manage_licensesForm'},
		{'label':'Categories', 'action':'manage_categoriesForm'},
		{'label':'Statuses', 'action':'manage_statusesForm'},
		{'label':'Databases', 'action':'manage_databaseForm'},
		{'label':'Operating Systems', 'action':'manage_operatingSystemsForm'},
		{'label':'CPUS', 'action':'manage_CPUForm'},
		{'label':'Updater', 'action':'manage_updateForm'},
		)
	
	#
	# XXX These should be ttw editable entities in the final system
	#
	softwareLicenses=[
		{'tag':'OSI', 'label':'OSI Approved'},
		{'tag':'apache', 'label':'Apache Software License'},
		{'tag':'apple', 'label':'Apple Public Source License'},
		{'tag':'artistic', 'label':'Artistic License'},
		{'tag':'bsdl', 'label':'BSD License'},
		{'tag':'gpl', 'label':'GNU General Public License'},
		{'tag':'lgpl', 'label':'GNU Lesser General Public License'},
		{'tag':'ibmc', 'label':'IBM Common Public License'},
		{'tag':'ibm', 'label':'IBM Public License'},
		{'tag':'intel', 'label':'Intel Open Source License'},
		{'tag':'jabber', 'label':'Jabber Open Source License'},
		{'tag':'mit', 'label':'MIT License'},
		{'tag':'mitre', 'label':'MITRE Collaborative Workspace License'},
		{'tag':'mpl', 'label':'Mozilla Public License'},
		{'tag':'mpl10', 'label':'Mozilla Public License 1.0'},
		{'tag':'mpl11', 'label':'Mozilla Public License 1.1'},
		{'tag':'nethack', 'label':'Nethack Open Source License'},
		{'tag':'nokia', 'label':'Nokia Open Source License'},
		{'tag':'python', 'label':'Python License'},
		{'tag':'qpl', 'label':'QT Public License'},
		{'tag':'ricoh', 'label':'Ricoh Source Code Public License'},
		{'tag':'sleepycat', 'label':'Sleepycat License'},
		{'tag':'sun', 'label':'Sun Internet Standards Source License'},
		{'tag':'vovida', 'label':'Vovida Software License 1.0'},
		{'tag':'zlib', 'label':'zlib or libpng License'},
		{'tag':'commercial', 'label':'Commercial Software'},
		{'tag':'freepu', 'label':'Free for Public Use'},
		{'tag':'pd', 'label':'Public Domain (no copyright)'},
		]
	
	#
	# XXX These should be ttw editable entities in the final system
	#
	productCategories=['External Access',
					   'Database Adapter',
					   'Server',
					   'Navigational',
					   'Helpers',
					   'Content Objects',
					   'Feedback',
					   'User Management',
					   'Commerce',
					   'Visual',
					   'Internationalisation',
					   'Misc']
	#
	# XXX These should be ttw editable entities in the final system
	#
	productStatus=['Mature',
				   'Stable',
				   'Mostly Stable',
				   'Beta',
				   'Alpha',
				   'Experimental',
				   'Preview',
				   'Obsolete',
				   ]

	# As above
	operatingSystems=[
	{'tag':'any', 'label':'Any'},
	{'tag':'unix', 'label':'Any UNIX like'},
	{'tag':'windows', 'label':'Any Windows'},
	{'tag':'freebsd', 'label':'FreeBSD'},
	{'tag':'linux', 'label':'Linux'},
	{'tag':'solaris', 'label':'Solaris'},
	{'tag':'unixware', 'label':'UnixWare'},
	{'tag':'aix', 'label':'AIX'},
	{'tag':'hpux', 'label':'HPUX'},
	{'tag':'qnx', 'label':'QNX'},
	{'tag':'netwware', 'label':'NetWare'},
	{'tag':'win95', 'label':'Windows 95'},
	{'tag':'win98', 'label':'Windows 98'},
	{'tag':'wince', 'label':'Windows CE'},
	{'tag':'winme', 'label':'Windows ME'},
	{'tag':'winnt', 'label':'Windows NT'},
	{'tag':'winxp', 'label':'Windows XP'},
		]

	# Yeah yeah I'm bad
	cpus=[
		'any',
		'i386',
		'ia64',
		'alpha',
		'ppc',
		'mips',
		'sparc',
		'ultrasparc',
		'other',
		]

	# I HAD TO START SOMEWHERE!
	productDatabases=[
		{'tag':'any', 'label':'Any'},
		{'tag':'odbc', 'label':'Any ODBC'},
		{'tag':'postgres', 'label':'Postgres'},
		{'tag':'interbase', 'label':'Interbase / Firebird'},
		{'tag':'oracle', 'label':'Oracle'},
		{'tag':'db2', 'label':'IBM DB2'},
		{'tag':'gadfly', 'label':'Gadfly'},
		{'tag':'ingres', 'label':'Ingres'},
		{'tag':'sybase', 'label':'Sybase'},
		{'tag':'mysql', 'label':'MySQL'},
		{'tag':'mssql', 'label':'Microsoft SQL Server'},
		{'tag':'access', 'label':'Microsoft Access'},
		{'tag':'sap', 'label':'SAP'},
		{'tag':'dbmaker', 'label':'DBMaker'},
	]
	installedPackages=[]
	remotePackageManagers=[]
	ignoreProducts=[]
	
	def __init__(self, id, title=''):
		self.id=id
		self.title=title
		self.packages=OOBTree()
		for c in self.productCategories:
			self.packages[c]=OOBTree()
		
	def reqattr(self, request, attr):
		try:    return request[attr]
		except: return None

	def updatePackage(self, package):
		self.packages[package.category][package.name+':'+package.version]=package
	def addPackage(self, package):
		self.packages[package.category][package.name+':'+package.version]=package

	def manage_createProduct(self, REQUEST):
		""" """
		package=ProductPackageMetadata(REQUEST)
		self.addPackage(package)
		return self.manage_viewPackageList(self, REQUEST)

		
	def manage_editPackage(self, name, version, category, REQUEST):
		""" Edit a package """
		package=self.getOnePackage(name, version, category)
		package.editDetails(REQUEST)
		if name != REQUEST.new_name or version != REQUEST.new_version:
			del self.packages[category][name+':'+version]
			name=REQUEST.new_name
			version=REQUEST.new_version
			REQUEST.version=REQUEST.new_version
			REQUEST.name=REQUEST.new_name
		self.updatePackage(package)
		return self.manage_editPackageForm(self, REQUEST)

	def manage_editPackageArchitecture(self, name, version, category, REQUEST):
		""" Edit Package Architecture """
		package = self.getOnePackage(name, version, category)
		package.editArchitecture(REQUEST)
		self.updatePackage(package)
		return self.manage_editArchitectureForm(self, REQUEST)

	def manage_editPackageDatabaseRequirements(self, name, version, category, REQUEST):
		""" Edit Database Requirements """
		package = self.getOnePackage(name, version, category)
		package.editDatabase(REQUEST)
		self.updatePackage(package)
		return self.manage_editDatabaseRequirementsForm(self, REQUEST)

	def manage_editPackageClassification(self, name, version, category, REQUEST):
		""" Edit Package Classication """
		package = self.getOnePackage(name, version, category)
		package.editClassification(REQUEST)
		if category != REQUEST.product_category:
			del self.packages[category][name+':'+version]
		self.updatePackage(package)
		return self.manage_editClassificationForm(self, REQUEST)

	def manage_editPackageDownloadSites(self, name, version, category, REQUEST):
		""" Edit Download Sites """
		package = self.getOnePackage(name, version, category)
		package.editDownloadSites(REQUEST)
		self.updatePackage(package)
		return self.manage_editDownloadSitesForm(self, REQUEST)

	def manage_editPackageDependencies(self, name, version, category, REQUEST):
		""" Edit Dependencies """
		package = self.getOnePackage(name, version, category)
		package.editDependencies(REQUEST)
		self.updatePackage(package)
		return self.manage_editDependenciesForm(self, REQUEST)
	
	def getAllPackages(self):
		pp = []
		for c in self.packages.keys():
			for p in self.packages[c].items():
				pp.append(p)
		return pp

	def getCategories(self):
		""" Get all Categories """
		return self.packages.items()

	def getCategoryCount(self, category):
		return len(self.packages[category].keys())

	def getCategoryPackages(self, category):
		""" Return all Packages for a Category """
		return self.packages[category].items()

	def getOnePackage(self, name, version, category):
		p = None
		try:
			p = self.packages[category][name+':'+version]
		except KeyError:
			pass
		return p

	# Wrapped in a list so we can use dtml-in which makes the dtml
	# cleaner to deal with...
	def getPackage(self, name, version, category):
		""" get a package by name and version """
		p = None
		try:
			p = self.packages[category][name+':'+version]
		except KeyError:
			pass
		if p:
			return [p]

	def ignoreControlPanelProduct(self, id):
		return id in self.ignoreProducts

	def controlPanelActions(self, ids, REQUEST):
		""" Perform actions on the ids returned from the Control Panel """
		if REQUEST.has_key('ignore'):
			for id in ids:
				self.ignoreProducts.append(id)
			
		return self.manage_updateForm(self, REQUEST)
	
	def markAsInstalled(self, package):
		packageInst={'name':package.name,
					 'version':package.version,
					 'category':package.category,
					 'installed_at':time.time()}

		self.installedPackages.append(packageInst)

	# Return true if v1 is probably greater than v2
	def versionIsProbablyGreater(self, v1, v2):
		vl1 = string.split(string.strip(v1), '.')
		vl2 = string.split(string.strip(v2), '.')
		if len(vl1) == len(vl2): # OK these things are the same
			try:
				ivl1 = map(int, vl1)
				ivl2 = map(int, vl2)
			except:
				# We can't compare, or they're not all ints
				# Return plain string compare
				return (v2 > v1)
			for i in range(len(ivl1)):
				if ivl2[i] > ivl1[i]:
					return 1
				if ivl2[i] == ivl1[i]:
					continue
				break
			else:
				return 1
		return 0
		
	def upgradeIsAvailable(self, category, name, version):
		treeKey=name+':'+version
		nameLen=len(name)
		currentVersion=''
		for pk in self.packages[category].keys():
			if ( pk[:nameLen] == name and
				 pk > treeKey ):
				# 0.2.0 should be < 0.10.0
				if self.versionIsProbablyGreater(version, pk[nameLen+1:]):
					if (( not currentVersion) or
						(currentVersion and self.versionIsProbablyGreater(currentVersion, pk[nameLen+1:]))):
						currentVersion=pk[nameLen+1:]

		return currentVersion
	
	def getInstalledPackages(self):
		""" Get installed Packages """
		return Results.Results( (({'name':'category', 'type':'s'},
								  {'name':'name', 'type':'s'},
								  {'name':'version', 'type':'version'}),
								 self.installedPackages))
		# return self.installedPackages

	def isInstalled(self, category, version, name):
		for p in self.installedPackages:
			if ( p['category'] == category and
				 p['name'] == name and
				 p['version'] == version ):
				return 1
		return 0
	
	def exportPackageList(self, exportCategories, exportLicenses,
						  exportStatuses, exportDatabases, exportPlatforms,
						  categories, exportTo, REQUEST={}):
		""" Call this get the package list """
		l = PackageListMetadata()
		if exportCategories:
			l.categories=self.getProductCategories()
		if exportLicenses:
			l.licenses=self.getSoftwareLicenses_internal()
		if exportStatuses:
			l.status=self.getProductStatuses()
		if exportDatabases:
			l.databases=self.getProductDatabases_internal()
		if exportPlatforms:
			l.operatingSystems=self.getOperatingSystems_internal()
			l.cpus=self.getCPUs()

		l.packages=[]
		for key in categories:
			category=self.packages[key]
			for pkey, product in category.items():
				l.packages.append(product.exportObject())

		fname=''
		if exportTo=='browser':
			REQUEST.RESPONSE.setHeader('Content-type', 'application/data')
			REQUEST.RESPONSE.setHeader('Content-Disposition',
									   'inline;filename=packagelist.xml')
			return l.dumps()
		else: # server
			fname = os.path.join(CLIENT_HOME, 'packagelist.xml')
			file=open(fname, 'w+b')
			file.write(l.dumps())
			file.close()

		if REQUEST:
			return self.manage_viewPackageList(self, REQUEST,
											   manage_tabs_message=
											   'Sucessfully exported to %s'%(fname),
											   title = 'Object exported')
		return ''

	# This needs to be smarter about merging the stuff in
	# We'll just be ultra stupid for now, we'll put in creation/modification
	# dates to enable partial updates.
	def importStuffFromXMLPickle(self, pl):
		cat_add=0
		lic_add=0
		stat_add=0
		pack_add=0
		db_add=0
		cpu_add=0
		os_add=0
		
		if hasattr(pl, 'categories'):
			if pl.categories:
				categories=self.getProductCategories()
				for pc in pl.categories:
					if pc not in categories:
						self.manage_addCategory(pc)
						cat_add=cat_add+1

		if hasattr(pl, 'licenses'):
			if pl.licenses:
				licenses={}
				lic_map = self.getSoftwareLicenses_internal()
				for l in lic_map:
					licenses[l['tag']]=l['label']
				for pc in pl.licenses:
					if pc['tag'] not in licenses.keys():
						self.manage_addLicense(pc['tag'], pc['label'])
						lic_add=lic_add+1

		if hasattr(pl, 'status'):
			if pl.status:
				statuses=self.getProductStatuses()
				for pc in pl.status:
					if pc not in statuses:
						self.manage_addStatus(pc)
						stat_add=stat_add+1
						

		if hasattr(pl, 'databases'):
			if pl.databases:
				databases={}
				db_map=self.getProductDatabases_internal()
				for l in db_map:
					databases[l['tag']]=l['label']
				for pc in pl.databases:
					if pc['tag'] not in databases.keys():
						self.manage_addDatabase(pc['tag'], pc['label'])
						db_add=db_add+1
		if hasattr(pl, 'cpus'):
			if pl.cpus:
				for pc in pl.cpus:
					if pc not in self.getCPUs():
						self.manage_addCPU(pc)
						cpu_add=cpu_add+1
		if hasattr(pl, 'operatingSystems'):
			if pl.operatingSystems:
				oss={}
				os_map=self.getOperatingSystems_internal()
				for o in os_map:
					oss[o['tag']]=o['label']
				for pc in pl.operatingSystems:
					if pc['tag'] not in oss.keys():
						self.manage_addOperatingSystem(pc['tag'], pc['label'])
						os_add=os_add+1

						
		for pobj in pl.packages:
			p=pobj.py_obj
			if not self.getPackage(p.name, p.version, p.category):
				self.addPackage(metadataToProductPackage(p))
				pack_add=pack_add+1

		return({'categories':cat_add, 'licenses':lic_add, 'status':stat_add,
				'packages':pack_add, 'os':os_add, 'db':db_add, 'cpu':cpu_add})
		
	def manage_importDataFromFile(self, inputFile, REQUEST={}):
		""" Import data directly from a file """
		start=time.time()
		xml_str=inputFile.read()
		pl = PackageListMetadata().loads(xml_str)
		added = self.importStuffFromXMLPickle(pl)
		end = time.time()
		if REQUEST:
			return self.manage_viewPackageList(self, REQUEST,
											   manage_tabs_message=
											   'Categories: %d Licenses: %d Statuses: %d Databases: %d Operating Systems: %d CPUS: %d New Packages: %d in %f seconds'%(
				added['categories'], added['licenses'], added['status'],
			    added['db'], added['os'], added['cpu'],
				added['packages'], end-start))
		return ''

	def manage_importDataFromRemote(self, remoteURL, remoteName='', REQUEST={}):
		""" Import data from a remote Product Manager """
		start=time.time()
		if remoteName and remoteURL:
			self.remotePackageManagers.append({'remoteName':remoteName, 'remoteURL':remoteURL})
		fname, header=urllib.urlretrieve(remoteURL)
		pn = open(fname, 'r')
		xml_str=pn.read()
		pn.close()
		pl = PackageListMetadata().loads(xml_str)
		os.unlink(fname)
		added = self.importStuffFromXMLPickle(pl)
		end = time.time()
		if REQUEST:
			return self.manage_viewPackageList(self, REQUEST,
											   manage_tabs_message=
											   'Categories: %d Licenses: %d Statuses: %d Databases: %d Operating Systems: %d CPUS: %d New Packages: %d in %f seconds'%(
				added['categories'], added['licenses'], added['status'],
			    added['db'], added['os'], added['cpu'],
				added['packages'], end-start))

		return ''
				
	def manage_exportPackage(self, category, version, name, REQUEST={}):
		""" Export a package definition """
		package = self.getOnePackage(name, version, category)
		xml_str = package.exportXML()
		if REQUEST:
			REQUEST.RESPONSE.setHeader('Content-Type','text/xml')
		return xml_str
		
	def manage_installPackages(self, packs, REQUEST={}):
		""" INSTALL A LIST OF PRODUCTS """
		for pack in packs:
			category, name, version = string.split(pack, ':')
			self.manage_installPackage(category, version, name, REQUEST)

	def getInstalledVersion(self, category, name):
		currentVersion=''
		for p in self.installedPackages:
			if p['category'] == category and p['name'] == name:
				if p['version'] > currentVersion:
					if self.versionIsProbablyGreater(currentVersion, p['version']):
						currentVersion=p['version']
		return currentVersion
	#
	# WARNING NOT MT-SAFE (yet)
	#
	# PROBABLY ALSO NOT SAFE FOR UPGRADING vs INSTALLING
	#
	def manage_installPackage(self, category, version, name, REQUEST={}):
		""" INSTALL A PRODUCT """
		installed_ok=0
		package = self.getOnePackage(name, version, category)
		if package:
			for dep in package.getDependencies():
				d_category, d_name, d_version = string.split(dep, ':')
				#
				# Don't install if it's installed
				#
				if isInstalled(d_category, d_version, d_name):
					continue

				#
				# Don't install if a later version is installed..
				#
				installedVersion = self.getInstalledVersion(d_name, d_category)

				if installedVersion:
					if installedVersion > d_version:
						if self.versionIsProbablyGreater(installedVersion, d_version):
							continue

				# Install this dependency and all of its dependencies
				self.manage_installPackage(d_category, d_version, d_name)
			# We assign to self so that Acquisition works...
			bleh = '_v_install_%s_%s_%s'%(category, version, name)
			setattr(self, bleh, ProductInstaller(category, version, name))

			_v_cccp = getattr(self, bleh)
			installed_ok = _v_cccp.install()
			del _v_cccp
			if installed_ok:
				self.markAsInstalled(package)
				#
				# Attempt to install it
				#
				bleh = get_products()
				# (priority, name, i, product_dir)
				for g_p, g_name, g_i, g_p_d in bleh:
					if name == g_name:
						meta_types=[]
						install_product(self.getPhysicalRoot(), g_p_d, g_name, meta_types,
										get_folder_permissions())
						add_mt=[]
						for new_mt in meta_types:
							if new_mt not in self.Control_Panel.meta_types:
								add_mt.append(new_mt)
						if add_mt:
							self.Control_Panel.meta_types=self.Control_Panel.meta_types+tuple(add_mt)
				
			else:
				if REQUEST:
					raise 'InstallationError', 'Product: %s, Version: %s'%(name, version)
				else:
					raise 'InstallationError', 'Product: %s, Version: %s (possibly being installed as a dependency)'%(name, version)
				
		if REQUEST:
			return self.manage_updateForm(self, REQUEST)
		return installed_ok

	manage_upgradePackage=manage_installPackage
	
	def convertZopeVersion(self, text):
		if not text:
			return '0.0.0' # some people don't deserve versions d8)
		a= string.replace(string.strip(string.join(string.split(text, '-')[1:])), ' ', '.')
		if a:
			return a
		a= string.replace(string.strip(string.split(text, ' ')[-1]), ' ', '.')
		if a:
			return a
		return text
	
	def getInstalledProductsList(self):
		ipl=[]
		for cp in self.Control_Panel.Products.objectIds():
			ipl.append(getattr(self.Control_Panel.Products, cp))
		self.installControlPanelProducts(ipl)
		return ipl

	def findControlPanelProduct(self, controlPanelProduct):
		name = controlPanelProduct.id
		version=self.convertZopeVersion(controlPanelProduct.version)
		treeKey=name+':'+version
		for c in self.packages.keys():
			if self.packages[c].has_key(treeKey):
				return(self.packages[c][treeKey])
					   
	def installControlPanelProducts(self, productList):
		for p in productList:
			rp = self.findControlPanelProduct(p)
			if rp:
				if not self.isInstalled(rp.category, rp.version, rp.name):
					self.markAsInstalled(rp)
				self.ignoreProducts.append(p.id)


	def getProductDatabases_internal(self):
		return self.productDatabases

	def getOperatingSystems_internal(self):
		return self.operatingSystems

	def getSoftwareLicenses_internal(self):
		return self.softwareLicenses

	def getProductCategories(self):
		return self.productCategories

	def getProductStatuses(self):
		return self.productStatus

	def getCPUs(self):
		return self.cpus

	# Pretend they came from an sql query (avoid 'mapping' keyword)
	def getProductDatabases(self):
		return Results.Results( (({'name':'tag', 'type':'s'},{'name':'label', 'type':'s'}), self.productDatabases)) 		

	# Pretend they came from an sql query (avoid 'mapping' keyword)
	def getOperatingSystems(self):

		return Results.Results( (({'name':'tag', 'type':'s'},{'name':'label', 'type':'s'}), self.operatingSystems)) 

	# Pretend they came from an sql query (avoid 'mapping' keyword)
	def getSoftwareLicenses(self):
		return Results.Results( (({'name':'tag', 'type':'s'},{'name':'label', 'type':'s'}), self.softwareLicenses))


	def manage_addLicense(self, tag, name, REQUEST={}):
		""" Add a license to the list """
		self.softwareLicenses.append({'tag':tag, 'label':name})
		if REQUEST:
			return self.manage_licensesForm(self, REQUEST)
		return ''

	def manage_addCategory(self, category, REQUEST={}):
		""" Add a category to the list """
		if category not in self.productCategories:
			self.productCategories.append(category)
			self.packages[category]=OOBTree()

		if REQUEST:
			return self.manage_categoriesForm(self, REQUEST)
		return ''

	def manage_addStatus(self, status, REQUEST={}):
		""" Add a status to the list """
		if status not in self.productStatus:
			self.productStatus.append(status)

		if REQUEST:
			return self.manage_statusForm(self, REQUEST)
		return ''

	def manage_addDatabase(self, tag, name, REQUEST={}):
		""" Add a Database to the List """
		self.productDatabases.append({'tag':tag, 'label':name})
		if REQUEST:
			return self.manage_databaseForm(self, REQUEST)
		return ''

	def manage_addOperatingSystem(self, tag, name, REQUEST={}):
		""" Add a CPU to the List """
		self.operatingSystems.append({'tag':tag, 'label':name})
		if REQUEST:
			return self.manage_operatingSystemsForm(self, REQUEST)
		return ''

	def manage_addCPU(self, cpu):
		self.cpus.append(cpu)
		if REQUEST:
			return self.manage_CPUForm(self, REQUEST)
		return ''

	
class ProductManagerRDBMS(ProductManager):
	""" A Product Manager Product which stores the data in a db"""
	meta_type='Product Manager RDBMS'

	def __init__(self, id, connection, title=''):
		self.id=id
		self.title=title
		self.connection=connection
		self.addSQLQueries()

	def addSQLQueries(self):
		self.sqlInsertLicense=SQL(
			'sqlInsertLicense',
			'Insert a New Software License',
			self.connection,
			'tag label',
			_sqlInsertLicense)
		self.sqlInsertProductDatabase=SQL(
			'sqlInsertProductDatabase',
			'Insert a Database that can be used by a product',
			self.connection,
			'tag label',
			_sqlInsertProductDatabase)
		self.sqlInsertOperatingSystem=SQL(
			'sqlInsertOperatingSystem',
			'Insert a New Operating System',
			self.connection,
			'tag label',
			_sqlInsertOperatingSystem)
		self.sqlInsertCategory=SQL(
			'sqlInsertCategory',
			'Insert a New Category',
			self.connection,
			'categoryName',
			_sqlInsertCategory)
		self.sqlInsertCPU=SQL(
			'sqlInsertCPU',
			'Insert a New CPU',
			self.connection,
			'CPUName',
			_sqlInsertCPU)
		self.sqlInsertStatus=SQL(
			'sqlInsertStatus',
			'Insert a New Status',
			self.connection,
			'statusName',
			_sqlInsertStatus)

		self.sqlGetProductDatabases=SQL(
			'sqlGetProductDatabases',
			'Get all Product Supported Databases',
			self.connection,
			'',
			_sqlGetProductDatabases)
		self.sqlGetOperatingSystems=SQL(
			'sqlGetOperatingSystems',
			'Get all Operating Systems',
			self.connection,
			'',
			_sqlGetOperatingSystems)
		self.sqlGetLicenses=SQL(
			'sqlGetLicenses',
			'Get All Licenses',
			self.connection,
			'',
			_sqlGetLicenses)
		self.sqlGetCategories=SQL(
			'sqlGetCategories',
			'Get All Categories',
			self.connection,
			'',
			_sqlGetCategories)
		self.sqlGetStatuses=SQL(
			'sqlGetStatuses',
			'Get All Statuses',
			self.connection,
			'',
			_sqlGetStatuses)
		self.sqlGetCPUS=SQL(
			'sqlGetCPUS',
			'Get All CPUS',
			self.connection,
			'',
			_sqlGetCPUS)

		self.sqlGetLastInsertId=SQL(
			'sqlGetLastInsertId',
			'Get the last postgres inserted Id',
			self.connection,
			'sequence',
			_sqlGetLastInsertId)

		self.sqlInsertPackage=SQL(
			'sqlInsertPackage',
			'Insert a Package',
			self.connection,
			'name version category author website description status type '
			'dbrequired platform architecture osversion license',
			_sqlInsertPackage)

		self.sqlUpdatePackage=SQL(
			'sqlUpdatePackage',
			'Update a Package',
			self.connection,
			'name version category author website description status type '
			'dbrequired platform architecture osversion license package_id',
			_sqlUpdatePackage)

		self.sqlInsertPackageDependencies=SQL(
			'sqlInsertPackageDependencies',
			'Insert Package Dependencies',
			self.connection,
			'package_id dependency_id',
			_sqlInsertPackageDependencies)

		self.sqlInsertPackageDatabases=SQL(
			'sqlInsertPackageDatabases',
			'Insert package databases',
			self.connection,
			'package_id database',
			_sqlInsertPackageDatabases)

		self.sqlInsertPackageDownloadSites=SQL(
			'sqlInsertPackageDownloadSites',
			'Insert package download sites',
			self.connection,
			'package_id url',
			_sqlInsertPackageDownloadSites)

		self.sqlInsertPackageCPU=SQL(
			'sqlInsertPackageCPU',
			'Insert a Package CPU',
			self.connection,
			'package_id cpu',
			_sqlInsertPackageCPU)

		self.sqlInsertPackagePlatform=SQL(
			'sqlInsertPackagePlatform',
			'Insert a Package Platform',
			self.connection,
			'package_id platform',
			_sqlInsertPackagePlatform)

		self.sqlFindPackageId=SQL(
			'sqlFindPackageId',
			'Find a Package Id',
			self.connection,
			'name version category',
			_sqlFindPackageId)

		self.sqlGetOnePackage=SQL(
			'sqlGetOnePackage',
			'Get One Package from an Id',
			self.connection,
			'package_id',
			_sqlGetOnePackage)

		self.sqlGetPackageDatabases=SQL(
			'sqlGetPackageDatabases',
			'Get Package Databases',
			self.connection,
			'package_id',
			_sqlGetPackageDatabases)

		self.sqlGetPackageDownloadSites=SQL(
			'sqlGetPackageDownloadSites',
			'Get Package Download Sites',
			self.connection,
			'package_id',
			_sqlGetPackageDownloadSites)

		self.sqlGetPackageCPUs=SQL(
			'sqlGetPackageCPUs',
			'Get Package CPUs',
			self.connection,
			'package_id',
			_sqlGetPackageCPUs)

		self.sqlGetPackagePlatforms=SQL(
			'sqlGetPackagePlatforms',
			'Get Package Platforms',
			self.connection,
			'package_id',
			_sqlGetPackagePlatforms)

		self.sqlGetPackageDependencyIds=SQL(
			'sqlGetPackageDependencyIds',
			'Get Dependent Packages',
			self.connection,
			'package_id',
			_sqlGetPackageDependencyIds)

		self.sqlGetPackageIds=SQL(
			'sqlGetPackageIds',
			'Get All Package Ids',
			self.connection,
			'',
			_sqlGetPackageIds)

		self.sqlGetInstalledFlag=SQL(
			'sqlGetInstalledFlag',
			'Get Installed Flag',
			self.connection,
			'package_id',
			_sqlGetInstalledFlag)

		self.sqlGetCategoryPackageIds=SQL(
			'sqlGetCategoryPackageIds',
			'Get Package Ids for a Category',
			self.connection,
			'category',
			_sqlGetCategoryPackageIds)

		self.sqlDeleteDownloadSites=SQL(
			'sqlDeleteDownloadSites',
			'Delete download sites for a package',
			self.connection,
			'package_id',
			_sqlDeleteDownloadSites)

		self.sqlDeletePackageCPUs=SQL(
			'sqlDeletePackageCPUs',
			'Delete Package CPUs',
			self.connection,
			'package_id',
			_sqlDeletePackageCPUs)

		self.sqlDeletePackagePlatforms=SQL(
			'sqlDeletePackagePlatforms',
			'Delete Package Platforms',
			self.connection,
			'package_id',
			_sqlDeletePackagePlatforms)

		self.sqlMarkAsInstalled=SQL(
			'sqlMarkAsInstalled',
			'Mark something as Installed',
			self.connection,
			'package_id',
			_sqlMarkAsInstalled)

		self.sqlGetInstalledPackageIds=SQL(
			'sqlGetInstalledPackages',
			'Get installed Packages',
			self.connection,
			'',
			_sqlGetInstalledPackageIds)

		self.sqlGetCategoryCount=SQL(
			'sqlGetCategoryCount',
			'Get Count for Category',
			self.connection,
			'category',
			_sqlGetCategoryCount)

		self.sqlFindControlPanelPackageIds=SQL(
			'sqlFindControlPanelPackageIds',
			'Get Control Panel Package Ids',
			self.connection,
			'name version',
			_sqlFindControlPanelPackageIds)

		self.sqlFindAllMatchingVersionIds=SQL(
			'sqlFindAllMatchingVersionIds',
			'Find Matching all versions for a Name and Category',
			self.connection,
			'name category',
			_sqlFindAllMatchingVersionIds)
		
	def manage_createProduct(self, REQUEST):
		""" """
		package=ProductPackageMetadata(REQUEST)
		self.addPackage(package)
		return self.manage_viewPackageList(self, REQUEST)

	def updatePackage(self, package, package_id=None):
		if not package_id:
			package_id=self.findPackageId(package.category, package.name,
										  package.version)
		self.sqlUpdatePackage(
			name=package.name,
			version=package.version,
			category=package.category,
			author=package.productAuthor,
			website=package.productWebsite,
			description=package.description,
			status=package.status,
			type=package.type,
			dbrequired=package.dbrequired,
			platform=package.platform,
			architecture=package.architecture,
			osversion=package.osVersion,
			license=package.license,
			package_id=package_id)
		
	def addPackage(self, package):
		self.sqlInsertPackage(
			name=package.name,
			version=(package.version or '0.0.0'),
			category=package.category,
			author=package.productAuthor,
			website=package.productWebsite,
			description=package.description,
			status=package.status,
			type=package.type,
			dbrequired=package.dbrequired,
			platform=package.platform,
			architecture=package.architecture,
			osversion=package.osVersion,
			license=package.license)

		package_id=self.getLastInsertId('packages_package_id_seq')
		for d in package.dependencies:
			if d:
				dep_id=self.getPackageIdFromDependencyString(d)
				self.sqlInsertPackageDependencies(package_id=package_id,
												  dep_id=dependency_id)

		if package.dbrequired != 'none':
			for d in package.databases:
				self.sqlInsertPackageDatabases(package_id=package_id,
											   database=d)

		for ds in package.download_sites:
			self.sqlInsertPackageDownloadSites(package_id=package_id,
											   url=ds)
		for cpu in package.architecture:
			self.sqlInsertPackageCPU(package_id=package_id,
									 cpu=cpu)

		for platform in package.platform:
			self.sqlInsertPackagePlatform(package_id=package_id,
										  platform=platform)

	def findPackageId(self, category, name, version):
		r = self.sqlFindPackageId(category=category, name=name, version=version)
		for d in r:
			return self.sqlattr(d, 'package_id')
		
	def getPackageIdFromDependencyString(self, depString):
		category, name, version=string.split(depString, ':')
		return self.findPackageId(category, name, version)

	def dependencyStringFromPackageId(self, package_id):
		p = self.packageIdToPackage(package_id, 0)
		return p.category+':'+p.name+':'+p.version
	
	def packageIdToPackage(self, package_id, doDeps=1):
		r = self.sqlGetOnePackage(package_id=package_id)
		for sqlp in r:
			sqlattr=self.sqlattr
			md={}
			md['name']=sqlattr(sqlp, 'name')
			md['version']=sqlattr(sqlp, 'version')
			md['product_category']=sqlattr(sqlp, 'category')
			md['author']=sqlattr(sqlp, 'author')
			md['webpage']=sqlattr(sqlp, 'website')
			md['productDescription']=sqlattr(sqlp, 'description')
			md['product_status']=sqlattr(sqlp, 'status')
			md['package_type']=sqlattr(sqlp, 'type')
			dbrequired=sqlattr(sqlp, 'dbrequired')
			md['dbrequired']=dbrequired
			md['osversion']=sqlattr(sqlp, 'osversion')
			md['productLicense']=sqlattr(sqlp, 'license')

			cpus = []
			dbr = self.sqlGetPackageCPUs(package_id=package_id)
			for sqldb in dbr:
				cpus.append(sqlattr(sqldb, 'cpu'))
				
			md['architecture']=cpus

			platforms=[]
			dbr = self.sqlGetPackagePlatforms(package_id=package_id)
			for sqldb in dbr:
				platforms.append(sqlattr(sqldb, 'platform'))

			md['platform']=platforms

			databases=[]
			if dbrequired != 'none':
				dbr = self.sqlGetPackageDatabases(package_id=package_id)
				for sqldb in dbr:
					databases.append(sqlattr(sqldb, 'database'))
			md['database']=databases
			urls=[]
			dbr = self.sqlGetPackageDownloadSites(package_id=package_id)
			for sqldb in dbr:
				urls.append(sqlattr(sqldb, 'download_site'))
			md['product_urls']=urls
			dependencies=[]
			if doDeps:
				r = self.sqlGetPackageDependencyIds(package_id=package_id)
				for dep in r:
					depString = self.dependencyStringFromPackageId(
						self.sqlattr(dep, 'dependency_id'))
					dependencies.append(depString)
			md['dependencies']=dependencies
			
			p = ProductPackageMetadata(md)
			return p
		
	def manage_editPackage(self, name, version, category, REQUEST):
		""" Edit a package """
		package_id=self.findPackageId(category, name, version)		
		package=self.getOnePackage(name, version, category)
		package.editDetails(REQUEST)
		if name != REQUEST.new_name or version != REQUEST.new_version:
			name=REQUEST.new_name
			version=REQUEST.new_version
			REQUEST.version=REQUEST.new_version
			REQUEST.name=REQUEST.new_name
		self.updatePackage(package, package_id)
		return self.manage_editPackageForm(self, REQUEST)

	def manage_editPackageArchitecture(self, name, version, category, REQUEST):
		""" Edit Package Architecture """
		package = self.getOnePackage(name, version, category)
		package.editArchitecture(REQUEST)
		self.updatePackage(package)
		self.updatePackageArchitecture(package)
		return self.manage_editArchitectureForm(self, REQUEST)	

	def manage_editPackageClassification(self, name, version, category, REQUEST):
		package=self.getOnePackage(name, version, category)
		package.editClassification(REQUES)
		self.updatePackage(package)
		return self.manage_editDownloadSitesForm(self, REQUEST)

	def manage_editPackageDatabaseRequirements(self, name, version, category, REQUEST):
		""" Edit Database Requirements """
		package = self.getOnePackage(name, version, category)
		package.editDatabase(REQUEST)
		self.updatePackage(package)
		self.updatePackageDatabases(package_id, package)		
		return self.manage_editDatabaseRequirementsForm(self, REQUEST)
	
	def manage_editPackageDownloadSites(self, name, version, category, REQUEST):
		""" Edit Download Sites """
		package_id=self.findPackageId(category, name, version)
		package=self.getOnePackage(name, version, category)
		package.editDownloadSites(REQUEST)
		self.updatePackage(package)
		self.updatePackageDownloadSites(package_id, package)
		return self.manage_editDownloadSitesForm(self, REQUEST)		

	def updatePackageDownloadSites(self, package_id, package):
		self.sqlDeleteDownloadSites(package_id=package_id)
		for ds in package.download_sites:
			self.sqlInsertPackageDownloadSites(package_id=package_id,
											   url=ds)

	def manage_editPackageDependencies(self, name, version, category, REQUEST):
		""" Edit Dependencies """
		package_id=self.findPackageId(category, name, version)		
		package = self.getOnePackage(name, version, category)
		package.editDependencies(REQUEST)
		self.updatePackage(package)
		self.updatePackageDependencies(package_id, package)		
		return self.manage_editDependenciesForm(self, REQUEST)

	def updatePackageDependencies(self, package_id, package):
		self.sqlDeletePackageDependencies(package_id=package_id)
		for pd in package.dependencies:
			dep_id=self.getPackageIdFromDependencyString(pd)
			self.sqlInsertPackageDependencies(package_id=package_id,
											  dependency_id=dep_id)

	def updatePackageArchitecture(self, package_id, package):
		self.sqlDeletePackageCPUS(package_id=package_id)
		for pd in package.architecture:
			self.sqlInsertPackageCPU(package_id=package_id,
									 cpu=pd)
		self.sqlDeletePackagePlatforms(package_id=package_id)
		for pd in package.platform:
			self.sqlInsertPackagePlatform(package_id=package_id,
										  platform=platform)

	def dbResultSetToPackageList(self, r):
		packages=[]
		for pid in r:
			p = self.packageIdToPackage(self.sqlattr(pid, 'package_id'))
			packages.append(p)
		return packages

	def getAllPackages(self):
		r = self.sqlGetPackageIds()
		return self.dbResultSetToPackageList(r)

	def getCategories(self):
		return self.sqlGetCategories()

	def getCategoryCount(self, category):
		r = self.sqlGetCategoryCount(category=category)
		for x in r:
			return self.sqlattr(x, 'category_count')

	def getCategoryPackages(self, category):
		r = self.sqlGetCategoryPackageIds(category=category)
		return self.dbResultSetToPackageList(r)

	def getOnePackage(self, name, version, category):
		package_id=self.findPackageId(category, name, version)

		if package_id:
			return self.packageIdToPackage(package_id)

	def getPackage(self, name, version, category):
		p = self.getOnePackage(name, version, category)
		if p:
			return [p]

	def markAsInstalled(self, package):
		package_id=self.findPackageId(version=package.version,
									  name=package.name,
									  category=package.category)
		self.sqlMarkAsInstalled(package_id=package_id)

	def isInstalled(self, category, version, name):
		package_id=self.findPackageId(version=version, name=name, category=category)
		r = self.sqlGetInstalledFlag(package_id=package_id)
		for x in r:
			return self.sqlattr(x, 'installed')

	def exportPackageList(self, exportCategories, exportLicenses,
						  exportStatuses, exportDatabases, exportPlatforms,
						  categories, exportTo, REQUEST={}):
		""" Call this get the package list """
		l = PackageListMetadata()
		if exportCategories:
			l.categories=self.getProductCategories()
		if exportLicenses:
			l.licenses=self.getSoftwareLicenses_internal()
		if exportStatuses:
			l.status=self.getProductStatuses()
		if exportDatabases:
			l.databases=self.getProductDatabases_internal()
		if exportPlatforms:
			l.operatingSystems=self.getOperatingSystems_internal()
			l.cpus = self.getCPUs()
			
		l.packages=[]
		for key in categories:
			for product in self.getCategoryPackages(key):
				l.packages.append(product.exportObject())

		fname=''
		if exportTo=='browser':
			REQUEST.RESPONSE.setHeader('Content-type', 'application/data')
			REQUEST.RESPONSE.setHeader('Content-Disposition',
									   'inline;filename=packagelist.xml')
			return l.dumps()
		else: # server
			fname = os.path.join(CLIENT_HOME, 'packagelist.xml')
			file=open(fname, 'w+b')
			file.write(l.dumps())
			file.close()

		if REQUEST:
			return self.manage_viewPackageList(self, REQUEST,
											   manage_tabs_message=
											   'Sucessfully exported to %s'%(fname),
											   title = 'Object exported')
		return ''

	def importStuffFromXMLPickle(self, pl):
		cat_add=0
		lic_add=0
		stat_add=0
		pack_add=0
		db_add=0
		cpu_add=0
		os_add=0
		
		if hasattr(pl, 'categories'):
			if pl.categories:
				categories=self.getProductCategories()
				for pc in pl.categories:
					if pc not in categories:
						self.manage_addCategory(pc)
						cat_add=cat_add+1

		if hasattr(pl, 'licenses'):
			if pl.licenses:
				licenses={}
				lic_map = self.getSoftwareLicenses_internal()
				for l in lic_map:
					licenses[l['tag']]=l['name']
				for pc in pl.licenses:
					if pc['tag'] not in licenses.keys():
						self.manage_addLicense(pc['tag'], pc['label'])
						lic_add=lic_add+1

		if hasattr(pl, 'status'):
			if pl.status:
				statuses=self.getProductStatuses()
				for pc in pl.status:
					if pc not in statuses:
						self.manage_addStatus(pc)
						stat_add=stat_add+1
						

		if hasattr(pl, 'databases'):
			if pl.databases:
				databases={}
				db_map=self.getProductDatabases_internal()
				for l in db_map:
					databases[l['tag']]=l['name']
				for pc in pl.databases:
					if pc['tag'] not in databases.keys():
						self.manage_addDatabase(pc['tag'], pc['label'])
						db_add=db_add+1
		if hasattr(pl, 'cpus'):
			if pl.cpus:
				for pc in pl.cpus:
					if pc not in self.getCPUs():
						self.manage_addCPU(pc)
						cpu_add=cpu_add+1
		if hasattr(pl, 'operatingSystems'):
			if pl.operatingSystems:
				oss={}
				os_map=self.getOperatingSystems_internal()
				for o in os_map:
					oss[o['tag']]=o['name']
				for pc in pl.operatingSystems:
					if pc['tag'] not in oss.keys():
						self.manage_addOperatingSystem(pc['tag'], pc['label'])
						os_add=os_add+1
							
		packages={}
		# Build a list of packages, so we can add depdencies first
		if hasattr(pl, 'packages'):
			for pobj in pl.packages:
				p=pobj.py_obj
				pp = metadataToProductPackage(p)
				packages[p.category+':'+p.name+':'+p.version]=pp

			for k in packages.keys():
				p = packages[k]
				if not self.getOnePackage(p.name, p.version, p.category):
					pack_add=pack_add+self.addPackagesWithDependencies(packages, p)

		return({'categories':cat_add, 'licenses':lic_add, 'status':stat_add,
				'packages':pack_add, 'os':os_add, 'db':db_add, 'cpu':cpu_add})

	def addPackagesWithDependencies(self, pl, package):
		numDone=0
		if package.dependencies:
			for pd in package.dependencies:
				if pl.has_key(pd):
					numDone=numDone+self.addPackagesWithDependencies(pl, pl[pd])
					del pl[pd]
		if not self.getPackage(package.name, package.version, package.category):
			self.addPackage(package)
		return numDone+1

	def getSoftwareLicenses(self):
		return self.sqlGetLicenses()

	def getProductCategories(self):
		categories=[]
		r =  self.sqlGetCategories()
		for x in r:
			categories.append(self.sqlattr(x, 'categoryname'))
		return categories

	def getProductStatuses(self):
		statuses=[]	
		r = self.sqlGetStatuses()
		for x in r:
			statuses.append(self.sqlattr(x, 'statusname'))
		return statuses

	def getSoftwareLicenses_internal(self):
		licenses=[]
		r = self.sqlGetLicenses()
		for x in r:
			licenses.append({'tag':self.sqlattr(x, 'tag'),
							 'label':self.sqlattr(x, 'label')})
		return licenses

	def getProductDatabases_internal(self):
		dbs=[]
		r = self.sqlGetProductDatabases()
		for x in r:
			dbs.append({'tag':self.sqlattr(x, 'tag'),
						'label':self.sqlattr(x, 'label')})
		return dbs

	def getOperatingSystems_internal(self):
		os=[]
		r = self.sqlGetOperatingSystems()
		for x in r:
			os.append({'tag':self.sqlattr(x, 'tag'),
							 'label':self.sqlattr(x, 'label')})
		return os
	
	def getProductDatabases(self):
		return self.sqlGetProductDatabases()

	def getOperatingSystems(self):
		return self.sqlGetOperatingSystems()

	def getCPUs(self):
		cpus = []
		r = self.sqlGetCPUS()
		for x in r:
			cpus.append(self.sqlattr(x, 'cpu'))
		return cpus

	def manage_addLicense(self, tag, name, REQUEST={}):
		""" Add a License """
		self.sqlInsertLicense(tag=tag,label=name)
		if REQUEST:
			return self.manage_licensesForm(self, REQUEST)
		return ''

	def manage_addCategory(self, category, REQUEST={}):
		""" Add a Category """
		self.sqlInsertCategory(categoryName=category)
		if REQUEST:
			return self.manage_categoriesForm(self, REQUEST)
		return ''

	def manage_addStatus(self, status, REQUEST={}):
		""" Add a Status """
		self.sqlInsertStatus(statusName=status)
		if REQUEST:
			return self.manage_statusForm(self, REQUEST)
		return ''

	def manage_addDatabase(self, tag, name, REQUEST={}):
		""" Add a Database to the List """
		self.sqlInsertProductDatabase(tag=tag, label=name)
		if REQUEST:
			return self.manage_databaseForm(self, REQUEST)
		return ''

	def manage_addOperatingSystem(self, tag, name, REQUEST={}):
		""" Add an Operating System """
		self.sqlInsertOperatingSystem(tag=tag, label=name)
		if REQUEST:
			return self.manage_operatingSystemsForm(self, REQUEST)
		return ''

	def manage_addCPU(self, cpu, REQUEST={}):
		""" Add a CPU """
		self.sqlInsertCPU(CPUName=cpu)
		if REQUEST:
			return self.manage_CPUForm(self, REQUEST)
		return ''

	def getLastInsertId(self, sequence):
		result=self.sqlGetLastInsertId(sequence=sequence)
		for x in result:
			return (self.sqlattr(x,'last_value'))
		
	def getInstalledPackages(self):
		r = self.sqlGetInstalledPackageIds()
		return self.dbResultSetToPackageList(r)

	def getInstalledVersion(self, category, name):
		currentVersion=''
		for p in self.getAllInstalledPackages():
			if p.category == category and p.name == name:
				if p.version > currentVersion:
					if self.versionIsProbablyGreater(currentVersion, p.version):
						currentVersion=p.version
		return currentVersion	

	def findAllMatchingVersions(self, category, name):
		r = self.sqlFindAllMatchingVersionIds(category=category, name=name)
		return self.dbResultSetToPackageList(r)
		
	def upgradeIsAvailable(self, category, name, version):
		currentVersion=''
		packages = self.findAllMatchingVersions(category, name)
		for pk in packages:
			if pk.version > version:
				if self.versionIsProbablyGreater(version, pk.version):
					if (( not currentVersion) or
						(currentVersion and self.versionIsProbablyGreater(currentVersion, pk.version))):
						currentVersion=pk.version

		return currentVersion	

	def findControlPanelProduct(self, controlPanelProduct):
		name = controlPanelProduct.id
		version=self.convertZopeVersion(controlPanelProduct.version)
		r = self.sqlFindControlPanelPackageIds(name=name,
											   version=version)
		return self.dbResultSetToPackageList(r)

	def installControlPanelProducts(self, productList):
		for p in productList:
			rpl = self.findControlPanelProduct(p)
			if rpl:
				for rp in rpl:
					if not self.isInstalled(rp.category, rp.version, rp.name):
						self.markAsInstalled(rp)
					self.ignoreProducts.append(p.id)

	def typeconv(self, val):
		if type(val)==mt:
			return ''
		return val

	def sqlattr(self, ob, attr):
		name=attr
		if hasattr(ob, attr):
			return self.typeconv(getattr(ob, attr))
		attr=upper(attr)
		if hasattr(ob, attr):
			return self.typeconv(getattr(ob, attr))
		attr=lower(attr)
		if hasattr(ob, attr):
			return self.typeconv(getattr(ob, attr))
		raise NameError, name

_sqlInsertLicense='''
INSERT INTO Licenses
VALUES(<dtml-sqlvar tag type=string>, <dtml-sqlvar label type=string>)
'''

_sqlInsertProductDatabase='''
INSERT INTO ProductDatabases
VALUES(<dtml-sqlvar tag type=string>, <dtml-sqlvar label type=string>)
'''

_sqlInsertOperatingSystem='''
INSERT INTO OperatingSystems
VALUES(<dtml-sqlvar tag type=string>, <dtml-sqlvar label type=string>)
'''

_sqlInsertCategory='''
INSERT INTO Categories VALUES (<dtml-sqlvar categoryName type=string>)
'''

_sqlInsertCPU='''
INSERT INTO CPUS VALUES (<dtml-sqlvar CPUName type=string>)
'''

_sqlInsertStatus='''
INSERT INTO Statuses VALUES(<dtml-sqlvar statusName type=string>)
'''

_sqlGetProductDatabases='''
SELECT * from ProductDatabases
'''

_sqlGetOperatingSystems='''
SELECT * from OperatingSystems
'''

_sqlGetLicenses='''
SELECT * FROM Licenses
'''

_sqlGetCategories='''
SELECT * FROM Categories
'''

_sqlGetStatuses='''
SELECT * FROM Statuses
'''

_sqlGetCPUS='''
SELECT * FROM CPUS
'''

_sqlGetLastInsertId='''
SELECT last_value FROM <dtml-var sequence>
'''

# packages_package_id_seq
_sqlUpdatePackage='''
UPDATE Packages
SET name=<dtml-sqlvar name type=nb>,
    version=<dtml-sqlvar version type=nb>,
	category=<dtml-sqlvar category type=nb>,
	author=<dtml-sqlvar author type=nb>,
	website=<dtml-sqlvar website type=string>,
	description=<dtml-sqlvar description type=string>,
	status=<dtml-sqlvar status type=nb>,
	type=<dtml-sqlvar type type=string>,
	dbrequired=<dtml-sqlvar dbrequired type=nb>,
	osversion=<dtml-sqlvar osversion type=nb>,
	license=<dtml-sqlvar license type=nb>,
	modified_date=NOW()
WHERE package_id=<dtml-sqlvar package_id type=int>
'''

_sqlInsertPackage='''
INSERT INTO Packages (name, version, category, author, website, description, status, type, dbrequired, osversion, license, creation_date, modified_date)
VALUES (<dtml-sqlvar name type=nb>,
        <dtml-sqlvar version type=nb>,
		<dtml-sqlvar category type=nb>,
		<dtml-sqlvar author type=nb>,
		<dtml-sqlvar website type=string>,
		<dtml-sqlvar description type=string>,
		<dtml-sqlvar status type=nb>,
		<dtml-sqlvar type type=string>,
		<dtml-sqlvar dbrequired type=nb>,
		<dtml-sqlvar osversion type=nb>,
		<dtml-sqlvar license type=nb>,
		NOW(), NOW())
'''

_sqlInsertPackageDependencies='''
INSERT INTO Package_dependencies VALUES(<dtml-sqlvar package_id type=int>,
                                       <dtml-sqlvar dependency_id type=int>)
'''

_sqlInsertPackageDatabases='''
INSERT INTO package_databases VALUES(<dtml-sqlvar package_id type=int>,
                                     <dtml-sqlvar database type=string>)
'''

_sqlInsertPackageDownloadSites='''
INSERT INTO package_download_sites VALUES(<dtml-sqlvar package_id type=int>,
                                          <dtml-sqlvar url type=string>)
'''

_sqlInsertPackageCPU='''
INSERT INTO package_cpus VALUES(<dtml-sqlvar package_id type=int>,
                                <dtml-sqlvar cpu type=string>)
'''

_sqlInsertPackagePlatform='''
INSERT INTO package_platforms VALUES(<dtml-sqlvar package_id type=int>,
	                                 <dtml-sqlvar platform type=string>)
'''

_sqlFindPackageId='''
SELECT package_id from packages
WHERE name=<dtml-sqlvar name type=string>
AND version=<dtml-sqlvar version type=string>
AND category=<dtml-sqlvar category type=string>
'''

_sqlGetCategoryPackageIds='''
SELECT package_id FROM packages
WHERE category=<dtml-sqlvar category type=nb>
ORDER BY name, version
'''

_sqlGetOnePackage='''
SELECT * FROM packages
WHERE package_id=<dtml-sqlvar package_id type=int>
'''

_sqlGetPackageDatabases='''
SELECT database FROM package_databases
WHERE package_id = <dtml-sqlvar package_id type=int>
'''

_sqlGetPackageDownloadSites='''
SELECT download_site FROM package_download_sites
WHERE package_id=<dtml-sqlvar package_id type=int>
'''

_sqlGetPackageCPUs='''
SELECT cpu from package_cpus
WHERE package_id=<dtml-sqlvar package_id type=int>
'''

_sqlGetPackagePlatforms='''
SELECT platform from package_platforms
WHERE package_id=<dtml-sqlvar package_id type=int>
'''

_sqlGetPackageDependencyIds='''
SELECT dependency_id FROM package_dependencies
WHERE package_id = <dtml-sqlvar package_id type=int>
'''

_sqlGetPackageIds='''
SELECT package_id from packages
ORDER BY creation_date, modified_date
'''

_sqlGetInstalledFlag='''
SELECT installed FROM packages
WHERE package_id = <dtml-sqlvar package_id type=int>
'''

_sqlDeleteDownloadSites='''
DELETE FROM package_download_sites
WHERE package_id = <dtml-sqlvar package_id type=int>
'''

_sqlDeletePackageCPUs='''
DELETE from package_cpus
WHERE package_id = <dtml-sqlvar package_id type=int>
'''

_sqlDeletePackagePlatforms='''
DELETE from package_platforms
WHERE package_id = <dtml-sqlvar package_id type=int>
'''

_sqlMarkAsInstalled='''
UPDATE packages set installed = 1
WHERE package_id = <dtml-sqlvar package_id type=int>
'''

_sqlGetInstalledPackageIds='''
SELECT package_id from packages
WHERE installed=1
ORDER BY creation_date, modified_date
'''

_sqlGetCategoryCount='''
SELECT COUNT(*) as category_count FROM Packages
WHERE category=<dtml-sqlvar category type=string>
'''

_sqlFindControlPanelPackageIds='''
SELECT package_id FROM Packages
WHERE name=<dtml-sqlvar name type=string>
AND version=<dtml-sqlvar version type=string>
'''

_sqlFindAllMatchingVersionIds='''
SELECT package_id FROM Packages
WHERE name=<dtml-sqlvar name type=string>
AND category=<dtml-sqlvar category type=string>
'''
