#
# Product Manager
#
# (C) Copyright 2000,2001 The Internet (Aust) Pty Ltd
# ACN: 082 081 472  ABN: 83 082 081 472
# All Rights Reserved
#
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
# OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
# OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
# SUCH DAMAGE.
#
# Author: Andrew Milton <akm@theinternet.com.au>

from xml_pickle import XML_Pickler

# List of Packages + categores, licenses etc
class PackageListMetadata(XML_Pickler):
	pass

# Individual Package
class PackageMetadata(XML_Pickler):
	pass
	
class ProductPackageMetadata:
	""" This describes the Product for listing in a master catalog """
	
	def __init__(self, md):
		# md is a dictionary (can be a REQUEST) that contains the following..
		# product_urls, product_category, product_status,
		# package_type, name, version, author, webpage, platform,
		# architecture, osversion, dbrequired, database, productDescription,
		# productLicense, productDependencies, externalDependencies

		self.name=md['name']
		self.version=md['version']
		self.category=md['product_category']		
		self.productAuthor=md['author']
		self.productWebsite=md['webpage']
		self.description=md['productDescription']
		self.status=md['product_status']
		self.type=md['package_type']

		self.download_sites=[]
		if md.has_key('product_urls'):
			pu=md['product_urls']
			for ds in pu:
				self.download_sites.append(ds)

		self.dbrequired=''
		self.databases=[]
		if md.has_key('dbrequired') and md['dbrequired']:
			self.dbrequired=md['dbrequired']
			if md.has_key('database'):
				dbs = md['database']
				for db in dbs:
					self.databases.append(db)

		self.dependencies=[]
		if md.has_key('dependencies'):
			self.dependencies=md['dependencies']
			
		self.platform=md['platform']
		self.architecture=md['architecture']
		self.osVersion=md['osversion']
		self.license=md['productLicense']

	def getDependencies(self):
		return self.dependencies
	
	def editDetails(self, md):
		self.name=md['new_name']
		self.version=md['new_version']
		self.description=md['productDescription']
		self.productAuthor=md['author']
		self.productWebsite=md['webpage']
		self.license=md['productLicense']

	def editArchitecture(self, md):
		self.platform=md['platform']
		self.architecture=md['architecture']
		self.osVersion=md['osversion']

	def editDatabase(self, md):
		self.dbrequired=md['dbrequired']
		self.databases=[]
		dbs = md['database']
		for db in dbs:
			self.databases.append(db)

	def editClassification(self, md):
		self.category=md['product_category']
		self.status=md['product_status']

	def editDownloadSites(self, md):
		self.download_sites=[]
		if md.has_key('product_urls'):
			pu=md['product_urls']
			for ds in pu:
				self.download_sites.append(ds)

	def editDependencies(self, md):
		self.dependencies=[]
		if md.has_key('dependencies'):
			self.dependencies=md['dependencies']	

	def getPackageMetaData(self):
		metadata=PackageMetadata()
		metadata.name=self.name
		metadata.version=self.version
		metadata.category=self.category		
		metadata.productAuthor=self.productAuthor
		metadata.productWebsite=self.productWebsite
		metadata.description=self.description
		metadata.status=self.status
		metadata.type=self.type
		metadata.download_sites=self.download_sites
		metadata.dbrequired=self.dbrequired
		metadata.databases=self.databases
		metadata.dependencies=self.dependencies
		metadata.platform=self.platform
		metadata.architecture=self.architecture
		metadata.osVersion=self.osVersion
		metadata.license=self.license
		return metadata

	def exportXML(self):
		metadata=self.getPackageMetaData()
		return metadata.dumps()

	def exportObject(self):
		return self.getPackageMetaData()

def metadataToProductPackage(pml):
	md = {}
	md['name']=pml.name
	md['version']=pml.version
	md['product_category']=pml.category
	md['author']=pml.productAuthor
	md['webpage']=pml.productWebsite
	md['productDescription']=pml.description
	md['product_status']=pml.status
	md['package_type']=pml.type
	md['product_urls']=pml.download_sites
	md['dbrequired']=pml.dbrequired
	md['database']=pml.databases
	md['dependencies']=pml.dependencies
	md['platform']=pml.platform
	md['architecture']=pml.architecture
	md['osversion']=pml.osVersion
	md['productLicense']=pml.license
	return ProductPackageMetadata(md)

class ProductPackageDependency:
	def __init__(self, minVersion, suggestedVersion, maxVersion):
		# all three are ProductPackageMetadata
		self.minVersion=minVersion
		self.maxVersion=maxVersion
		self.suggestedVersion=suggestedVersion


class ProductPackage:
	""" This is the actual package """
	def __init__(self, metadata, checksum, archive):
		# archive will be a zip file with the code in it.
		# metadata is a ProductPackageMetadata object
		# checksum is an md5 checksum of the archive.
		self.metadata=metadata
		self.checksum=checksum
		self.archive=archive
